import {
  useCallback, useContext, useEffect, useRef, useState,
} from 'react';
import { Web3 } from 'web3';

import Button from 'react-bootstrap/esm/Button';
import Stack from 'react-bootstrap/esm/Stack';
import useList from '../utils/List';
import Web3Context from '../contexts/Web3Context';
import ProposalsArtifact from '../assets/Proposals.json';
import { getContractAddressByChainId } from '../utils/contract-config';
import AddPostModal from './posts/AddPostModal';

const getProposalStatus = (proposal) => {
  switch (Number(proposal.stage)) {
    case 0: return 'Proposal';
    case 1: return 'Referendum 0%';
    case 2: return 'Referendum 1%';
    case 3: return 'Referendum 100%';
    case 4: return 'Failed';
    case 5: return 'Accepted';
    default: return 'Unknown';
  }
};

function Proposals() {
  const {
    provider, chainId, account, reputation,
  } = useContext(Web3Context);
  const [proposals, dispatchProposal] = useList();
  const proposalsContract = useRef();
  const [showAddProposal, setShowAddProposal] = useState(false);

  const fetchProposal = useCallback(async (proposalIndex) => {
    const proposal = await proposalsContract.current.methods.proposals(proposalIndex).call();
    proposal.pools = await proposalsContract.current.methods.getPools(proposalIndex).call();
    dispatchProposal({
      type: 'update',
      item: {
        ...proposal,
        id: proposalIndex,
      },

    });
    console.log('proposal.pools', proposal.pools);
  }, [proposalsContract, dispatchProposal]);

  const fetchProposals = useCallback(async () => {
    const count = await proposalsContract.current.methods.proposalCount().call();
    const promises = [];
    dispatchProposal({ type: 'refresh' });
    for (let i = 0; i < count; i += 1) {
      promises.push(fetchProposal(i));
    }
    await Promise.all(promises);
  }, [proposalsContract, fetchProposal, dispatchProposal]);

  // Initial data load and event subscription
  useEffect(() => {
    if (chainId === undefined) return () => {};
    const web3 = new Web3(provider);
    const ProposalsAddress = getContractAddressByChainId(chainId, 'Proposals');
    const contract = new web3.eth.Contract(ProposalsArtifact.abi, ProposalsAddress);
    proposalsContract.current = contract;

    fetchProposals();

    const onNewProposal = (event) => {
      console.log('event: new proposal', event.returnValues);
      fetchProposal(event.returnValues.proposalIndex);
    };

    const onAttestation = (event) => {
      console.log('event: attestation', event.returnValues);
      fetchProposal(event.returnValues.proposalIndex);
    };

    contract.events.NewProposal({ fromBlock: 'latest' }).on('data', onNewProposal);
    contract.events.Attestation({ fromBlock: 'latest' }).on('data', onAttestation);

    // unsubscribe on unmount
    return () => {
      contract.events.NewProposal().off(onNewProposal);
      contract.events.Attestation().off(onAttestation);
    };
  }, [provider, chainId, proposalsContract, fetchProposals, fetchProposal]);

  const handleShowAddProposal = () => setShowAddProposal(true);

  const onSubmitProposal = useCallback(async (post) => {
    // TODO: Make referenda durations configurable
    await proposalsContract.current.methods.propose(post.hash, 20, 40, 60).send({
      from: account,
      gas: 1000000,
      value: 10000,
    });
  }, [account, proposalsContract]);

  const handleAttest = useCallback(async (proposalIndex) => {
    await proposalsContract.current.methods.attest(proposalIndex, reputation).send({
      from: account,
      gas: 1000000,
    });
  }, [proposalsContract, account, reputation]);

  const handleEvaluateAttestation = useCallback(async (proposalIndex) => {
    await proposalsContract.current.methods.evaluateAttestation(proposalIndex).send({
      from: account,
      gas: 1000000,
    });
  }, [proposalsContract, account]);

  const getReferenda = (proposal) => {
    if (!proposal || !proposal.pools) return [];
    const referenda = [];
    for (let referendumIndex = 0; referendumIndex < 3; referendumIndex += 1) {
      for (let i = 0; i < 3; i += 1) {
        const pool = proposal.pools[referendumIndex][i];
        if (pool.started) {
          referenda.push(
            <div key={`${referendumIndex}.{i}`}>
              {`${referendumIndex}.${i}. `}
              {!pool.completed && (
              <span>
                In Progress
              </span>
              )}
              {pool.completed && (
              <span>
                {`${pool.stakedFor.toString()} U / ${pool.stakedAgainst.toString()} D`}
                {', '}
                {pool.votePasses ? 'Accepted' : 'Rejected'}
                {pool.quorumMet || ': Quorum not met'}
              </span>
              )}
            </div>,
          );
        }
      }
    }
    return referenda;
  };
  return (
    <>
      <AddPostModal title="New Proposal" show={showAddProposal} setShow={setShowAddProposal} onSubmit={onSubmitProposal} />
      <h2>Proposals</h2>
      <div>
        <Button onClick={handleShowAddProposal}>New Proposal</Button>
      </div>
      <div>
        <table className="table">
          <thead>
            <tr>
              <th>ID</th>
              <th>Fee</th>
              <th>Stage</th>
              <th>Attestation</th>
              <th>Referenda</th>
              <th>Actions</th>
            </tr>
          </thead>
          <tbody>
            {proposals.filter((x) => !!x).map((proposal) => (
              <tr key={proposal.id}>
                <td>{proposal.id.toString()}</td>
                <td>{proposal.fee.toString()}</td>
                <td>{getProposalStatus(proposal)}</td>
                <td>{proposal.attestationTotal.toString()}</td>
                <td>
                  <Stack direction="vertical">
                    {getReferenda(proposal)}
                  </Stack>
                </td>
                <td>
                  {proposal.stage === 0n && (
                    <>
                      <Button onClick={() => handleAttest(proposal.id)}>Attest</Button>
                      {' '}
                      <Button onClick={() => handleEvaluateAttestation(proposal.id)}>
                        Evaluate Attestation
                      </Button>
                    </>
                  )}
                </td>
              </tr>
            ))}
          </tbody>
        </table>
      </div>
    </>
  );
}

export default Proposals;
